/*
** game.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <game.h>

/*
** Initialize a game
**
** Arguments:
**  game		Structure containing all datas
**  argc		Arguments counter
**  argv		Arguments vector
**
** Return value:
**  -1			Error
**  0			Okay
*/
int			game_init(t_game *game, int argc, char **argv)
{
  unsigned		i;

  if (argc < 3)
    return (-1);
  game->res = strtoul(argv[1], 0, 10);
  game->l = argc - 2;
  game->n = malloc(game->l * sizeof (*game->n));
  game->u = malloc(game->l * sizeof (*game->u));
  for (i = 0; i < game->l; i++)
    {
      game->n[i] = strtoul(argv[2 + i], 0, 10);
      game->u[i] = 1;
    }
  game->ncalc = 0;
  game->calcs = malloc((game->l - 1) * sizeof (*game->calcs));
  game->trycnt = 0;
  return (0);
}

/*
** Compute the solutions
**
** Arguments:
**  game		Structure containing all datas
**  show_solution	Function pointer used to display a solution
**
** Return value:
**  Number of solutions found
*/
unsigned		game_compute(t_game *game,
				     void (*show_solution)(t_game *))
{
  unsigned		nsol;
  unsigned		i;
  unsigned		j;

  nsol = 0;
  if (-1 < game_res_reached(game))
    {
      nsol++;
      show_solution(game);
    }
  if (game_nb_used_numbers(game) < 2)
    return (nsol);
  for (i = 0; i < game->l - 1; i++)
    if (!game->u[i])
      continue ;
    else
      for (j = i + 1; j < game->l; j++)
	if (!game->u[j])
	  continue ;
	else
	  nsol += game_work_with_these_2nb(game, show_solution, i, j);
  return (nsol);
}

/*
** Explores the tree of calculations from the 2 nb
**
** Arguments:
**  game		Structure containing all datas
**  show_solution	Function pointer used to display a solution
**  n1_idx		Index in game->n of the 1st number choosed
**  n2_idx		Index in game->n of the 2nd number choosed
**
** Return value:
**  Number of solutions found
*/
unsigned	game_work_with_these_2nb(t_game *game,
					 void (*show_solution)(t_game *),
					 int n1_idx,
					 int n2_idx)
{
  unsigned		nsol = 0;
  int			n1;
  int			n2;

  nsol = 0;
  n1 = game->n[n1_idx];
  n2 = game->n[n2_idx];
  game->u[n2_idx] = 0;
  game->ncalc++;
  /* n1 op n2 */
  game->calcs[game->ncalc - 1].n1 = n1;
  game->calcs[game->ncalc - 1].n2 = n2;
  game->calcs[game->ncalc - 1].op = '+';
  game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n1 + n2;
  nsol += game_compute(game, show_solution);
  game->calcs[game->ncalc - 1].op = '*';
  game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n1 * n2;
  nsol += game_compute(game, show_solution);
  if (n1 - n2 >= 0)
    {
      game->calcs[game->ncalc - 1].op = '-';
      game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n1 - n2;
      nsol += game_compute(game, show_solution);
    }
  if (n2 && !(n1 % n2))
    {
      game->calcs[game->ncalc - 1].op = '/';
      game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n1 / n2;
      nsol += game_compute(game, show_solution);
    }
  /* n2 op n1 */
  game->calcs[game->ncalc - 1].n1 = n2;
  game->calcs[game->ncalc - 1].n2 = n1;
  if (n2 - n1 >= 0)
    {
      game->calcs[game->ncalc - 1].op = '-';
      game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n2 - n1;
      nsol += game_compute(game, show_solution);
    }
  if (n1 && !(n2 % n1))
    {
      game->calcs[game->ncalc - 1].op = '/';
      game->calcs[game->ncalc - 1].res = game->n[n1_idx] = n2 / n1;
      nsol += game_compute(game, show_solution);
    }
  /* End */
  game->ncalc--;
  game->n[n1_idx] = n1;
  game->u[n2_idx] = 1;
  return (nsol);
}

/*
** Determines whether the res is in the number list
**
** Arguments:
**  game		Structure containing all datas
**
** Return value:
**  -1			res is not in the list
**  >= 0		Index of res in the list
*/
int			game_res_reached(t_game *game)
{
  unsigned		i;

  game->trycnt++;
  for (i = 0; i < game->l; i++)
    if (game->u[i] && game->n[i] == game->res)
      return (i);
  return (-1);
}

/*
** Counts the number of used numbers (game->u[i] flag != 0) in the list
**
** Arguments:
**  game		Structure containing all datas
**
** Return value:
**  >= 0		Number of used numbers
*/
unsigned		game_nb_used_numbers(t_game *game)
{
  unsigned		i;
  unsigned		cnt;

  cnt = 0;
  for (i = 0; i < game->l; i++)
    cnt += !!game->u[i];
  return (cnt);
}

/*
** free()'s some elements in the game struct
**
** Arguments:
**  game		Structure containing all datas
*/
void			game_free(t_game *game)
{
  free(game->n);
  free(game->u);
  free(game->calcs);
}
